import numpy as np 
import pandas as pd
import functools
import collections
from abc import ABC, abstractmethod
from fcmpy.expert_fcm.input_validator import type_check
from fcmpy.store.methodsStore import EntropyStore
from fcmpy.store.methodsStore import ReaderStore
from fcmpy.store.methodsStore import MembershipStore
from fcmpy.store.methodsStore import ImplicationStore
from fcmpy.store.methodsStore import AggregationStore
from fcmpy.store.methodsStore import DefuzzStore
from fcmpy.expert_fcm.transform import Transform


class FcmConstructor(ABC):
    @abstractmethod
    def read_data(file_path, **kwargs) -> collections.OrderedDict:
        raise NotImplementedError('read_data method is not defined!')

    @abstractmethod
    def build(data: collections.OrderedDict, implication_method:str, 
                    aggregation_method:str, defuzz_method:str) -> pd.DataFrame:
                    
        raise NotImplementedError('build method is not defined!')


class ExpertFcm(FcmConstructor):
    """
        Construct Expert FCMs based on qualitative input data.

        Methods:
            read_data(file_path, **kwargs)
            entropy(data: collections.OrderedDict, method = 'entropy', **kwargs)
            automf(method:str='trimf', **kwargs)
            fuzzy_implication(membership_function, weight, method:str='Mamdani', **kwargs)
            aggregate(x, y, method:str='fMax', **kwargs)
            defuzz(x, mfx, method:str='centroid', **kwargs)
            build(data: collections.OrderedDict, implication_method:str='Mamdani', 
                            aggregation_method:str='fMax', defuzz_method:str='centroid')
    """
    @type_check
    def __init__(self):
        self.__linguisticTerms = None
        self.__membership = None
        self.__universe = None

    @property
    def linguistic_terms(self):
        return self.__linguisticTerms

    @linguistic_terms.setter
    @type_check
    def linguistic_terms(self, terms: dict):
        """
            Linguistic terms and the associated parameters for generating fuzzy membership functions.

            Parameters
            ----------
            terms : dict,
                        keys are the linguistic terms and the values are 
                        lists with term parameters (see blow)

            Term parameters
            ----------------
            for trimf:
                the parameters 'abc' should be passed as keys (in a list)
                to the linguistic terms -> e.g., {'+VL': [0.25, 0.5, 0.75]}
            
            for gaussmf:
                the parameters 'mean' and 'sigma' should be passed as keys (in a list)
                to the linguistic terms -> e.g., {'+VL': [0.25, 0.1]}
            
            for trapmf:
                the parameters 'abcd' should be passed as keys to the linguistic terms
                e.g., {'+VL': [0, 0.25, 0.5, 0.75]}
        """
        termsLower = dict((k.lower(), v) for k, v in terms.items()) 
        self.__linguisticTerms = termsLower

    @property
    def fuzzy_membership(self):
        return self.__membership
    
    @fuzzy_membership.setter
    @type_check
    def fuzzy_membership(self, membership:dict):   
        """
            Parameters
            ----------
            membership : dict,
                            keys are the linguistic terms and the values are the associated
                            numpy arrays with membership values.
        """

        self.__membership = membership

    @property
    def universe(self):
        return self.__universe
    
    @universe.setter
    @type_check
    def universe(self, universe:np.ndarray):
        
        """
        Parameters
        ----------
        universe: np.ndarray
                    the universe of discourse. 
                    (Note that the generated arrays are automatically rounded to 2 digits)
        """

        self.__universe = universe.round(2)

    @type_check
    def read_data(self, file_path:str, **kwargs) -> collections.OrderedDict:
        
        """ 
            Read data from a file. Currently, the method supports 
            .csv, .xlsx and .json file formats.
            
            Parameters
            ----------
            file_path : str 
            
            Other Parameters
            ----------------
            for .csv files:

                **sep_concept: str,
                                separation symbol (e.g., '->') that separates the antecedent 
                                from the consequent in the column heads of a csv file
                                default ---> '->'

                **csv_sep: str,
                            separator of the csv file
                            default ---> ','
                            
            for .xlsx files:

                **check_consistency: Bool
                                    check the consistency of raitings across the experts
                                    default --> False

                **engine: str,
                            the engine for excel reader (read more in pd.read_excel)
                            default --> "openpyxl"

            for .json files:

                **check_consistency: Bool
                                    check the consistency of raitings across the experts.
                                    default --> False
            
            Return
            -------
            data: collections.OrderedDict
                    ordered dictionary with the formatted data.
        """

        if self.linguistic_terms is None:
            raise ValueError('Linguistic terms are not defined!')

        fileType = file_path.split('.')[-1]
        reader = ReaderStore.get(fileType)()
        data = reader.read(filePath=file_path, linguisticTerms = self.linguistic_terms,
                            params = kwargs)
        
        return data

    @type_check
    def entropy(self, data: collections.OrderedDict, 
                method:str = 'entropy', **kwargs) -> pd.DataFrame:
        """
            Calculate the entropy of the expert ratings.

            Parameters
            ----------
            data: collections.OrderedDict
                ordered dictionary with the expert inputs
            
            method: str
                    method for calculating entropy. At the moment only information entropy is available
                    default --> 'entropy'

            Return
            -------
            y: pandas.DataFrame,
                entropy of the concept pairs in expert ratings
        """

        entropy = EntropyStore.get(method)()
        return entropy.calculateEntropy(data=data, params=kwargs)
    
    @type_check
    def automf(self, method:str='trimf', **kwargs) -> dict:
        """
            Generate membership functions for a given set of linguistic terms.

            Parameters
            ----------
            method: str
                    type of membership function. At the moment three such types are available  
                    'trimf': triangular membership functions
                    'gaussmf': gaussian membership functions
                    'trapmf': trapezoidal membership functions
                    default ---> 'trimf'
            
            Return
            -------
            y: dict
                generated membership functions. The keys are the linguistic
                terms and the values are 1d arrays
        """
        if self.linguistic_terms is None:
            raise ValueError('Linguistic terms are not defined!')
        elif self.universe is None:
            raise ValueError('Universe of discourse is not defined!')
        
        membership = MembershipStore.get(method)()

        return membership.membershipFunction(linguistic_terms=self.linguistic_terms, 
                                            universe=self.universe, params = kwargs)
    
    @type_check
    def fuzzy_implication(self, membership_function, weight, 
                            method:str='Mamdani', **kwargs) -> np.ndarray:
        """
            Fuzzy implication rule.

            Parameters
            ----------
            membership_function: numpy.ndarray,
                                    membership function of a linguistic term (x)

            weight: float,
                        the weight at which the given membership function x should be "activated" 
                        (i.e., the cut point or the point at which the membership function should be rescaled)
            
            method: str,
                    implication rule; at the moment two such rules are available 
                    'Mamdani': minimum implication rule
                    'Larsen': product implication rule
                    default ---> 'Mamdani'

            Return
            -------
            y: numpy.ndarray
                the "activated" membership function
        """
        implication = ImplicationStore.get(method)()

        return implication.implication(mf_x = membership_function, 
                                        weight = weight, params = kwargs)

    @type_check
    def aggregate(self, x, y, method:str='fMax', **kwargs) -> np.ndarray:
        """
            Fuzzy aggregation rule.

            Parameters
            ----------
            x, y: numpy.ndarray,
                        "activated" membership functions of the linguistic 
                        terms that need to be aggregated
            
            method: str,
                    aggregation rule; at the moment four such rules are available
                    'fMax': family maximum,
                    'algSum': family Algebraic Sum,
                    'eSum': family Einstein Sum,
                    'hSum': family Hamacher Sum
                    default ---> 'fMax'

            Return
            -------
            y: numpy.ndarray
                an aggregated membership function
        """

        aggr = AggregationStore.get(method)()
        aggregated = aggr.aggregate(x=x, y=y, params = kwargs)
        return aggregated

    @type_check
    def defuzz(self, x, mfx, method:str='centroid', **kwargs) -> float:
        """
            Defuzzification of the aggregated membership functions.

            Parameters
            ----------
            x: numpy.ndarray
                universe of discourse 
            
            mfx: numpy.ndarray,
                        "aggregated" membership functions
            
            method: str,
                    defuzzification method; at the moment four such 
                    rules are available:
                        'centroid': Centroid,
                        'bisector': Bisector,
                        'mom': MeanOfMax,
                        'som': MinOfMax,
                        'lom' : MaxOfMax
                        default ---> 'centroid'

            Return
            -------
            y: float
                defuzzified value
        """

        defuzz = DefuzzStore.get(method)()
        defuzz_value = defuzz.defuzz(x=x, mfx=mfx, method=method, params=kwargs)

        return defuzz_value

    @type_check
    def build(self, data: collections.OrderedDict, implication_method:str='Mamdani', 
                    aggregation_method:str='fMax', defuzz_method:str='centroid') -> pd.DataFrame:
        """
            Build an FCM based on qualitative input data.

            Parameters
            ----------
            data: collections.OrderedDict
                    ordered dictionary with the qualitative input data.
            
            implication_method: str,
                                implication rule; at the moment two such
                                rules are available;
                                    'Mamdani': minimum implication rule
                                    'Larsen': product implication rule
                                    default ---> 'Mamdani'

            aggregation_method: str,
                                aggregation rule; at the moment four such
                                rules are available:
                                    'fMax': family maximum,
                                    'algSum': family Algebraic Sum,
                                    'eSum': family Einstein Sum,
                                    'hSum': family Hamacher Sum
                                    default ---> 'fMax'

            defuzz_method: str,            
                            defuzzification method; at the moment four such
                            rules are available:
                                'centroid': Centroid,
                                'bisector': Bisector,
                                'mom': MeanOfMax,
                                'som': MinOfMax,
                                'lom' : MaxOfMax
                                default ---> 'centroid'

            Return
            -------
            y: pd.DataFrame
                the connection matrix with the defuzzified values            
        """
        if self.fuzzy_membership is None:
            raise ValueError('Membership function is not defined!')

        if self.linguistic_terms is None:
            raise ValueError('linguistic_terms are not defined!')

        if self.universe is None:
            raise ValueError('Universe of discourse is not defined!')

        nExperts = len(data)

        # Drop the columns that should be omitted from the calculations (e.g., unsure)
        keep = [i.lower() for i in list(self.linguistic_terms.keys())]
        flat_data = Transform.flatData(data)
        flat_data = flat_data[keep]
        
        # Create an empty weight matrix
        cols = set(flat_data.index.get_level_values('to'))
        index = set(flat_data.index.get_level_values('from'))
        index = sorted(index.union(cols))
        weight_matrix = pd.DataFrame(0,columns=index, index=index)
        # main part for calculating the weights
        for concepts in set(flat_data.index):
            # for a given pair of concepts calculate the propostions (weights) for the
            # implication rules.
            activation_parameter = Transform.calculateProportions(data=flat_data, 
                                            conceptPair=concepts, nExperts=nExperts)
            activated = {}
            # for each linguistic term apply the implication rule
            for term in self.fuzzy_membership.keys():
                act = self.fuzzy_implication(membership_function=self.fuzzy_membership[term], 
                                            weight=activation_parameter[term], method=implication_method)
                activated[term] = act
            
            # if the 'activated' membership functions are not all zeros then aggregate 
            # them and defuzzify them.
            if not all(x==0 for x in activation_parameter.values()):
                # aggregate all the activated membership functions
                aggregated = functools.reduce(lambda x,y: self.aggregate(x=x, y=y, method=aggregation_method),
                                                [activated[i] for i in activated.keys()])

                # defuzzify the aggregated functions                                
                value = self.defuzz(x=self.universe, mfx=aggregated, method=defuzz_method)
                # populate the empty weigtht_matrix with the defuzzified value
                weight_matrix.loc[concepts] = value
        
        weight_matrix = weight_matrix.fillna(0)

        return weight_matrix


import numpy as np
from abc import ABC, abstractclassmethod  
from fcmpy.expert_fcm.input_validator import type_check


class FuzzyImplication:
    """
        Fuzzy implication rules.
    """
    @abstractclassmethod
    def implication() -> np.ndarray:
        raise NotImplementedError('implication method is not defined.')


class Mamdani(FuzzyImplication):
    """
        Mamdani minimum implication rule.
    """
    @staticmethod
    @type_check
    def implication(**kwargs) -> np.ndarray: 
        """
            Mamdani min fuzzy implication rule.

            Other Parameters
            ----------------
            **mf_x: numpy.ndarray,
                    membership function of a linguistic term (x)

            **weight: float,
                        the weight at which the membership function x
                        should be activated (i.e., the cut point)
            
            Return
            -------
            y: numpy.ndarray
                the 'activated' membership function
        """
        mf_x = kwargs['mf_x']
        weight = kwargs['weight']

        return np.fmin(weight, mf_x)


class Larsen(FuzzyImplication):
    """
        Larsen's product implication rule.
    """
    @staticmethod
    @type_check
    def implication(**kwargs) -> np.ndarray:
        """
            Larsen's product fuzzy implication rule.

            Other Parameters
            ----------------
            **mf_x: numpy.ndarray
                    membership function of a linguistic term (x)

            **weight: float,
                        the weight at which the membership function x should be activated (i.e., rescaled)
            
            Return
            -------
            y: numpy.ndarray
                the activated (rescaled) membership function
        """
        mf_x = kwargs['mf_x']
        weight = kwargs['weight']

        return  np.dot(mf_x, weight)


import numpy as np
from abc import ABC, abstractmethod


class Inference(ABC):
    """
        Class of FCM inference methods
    """
    @abstractmethod
    def infer() -> np.array:
        raise NotImplementedError('Infer method is not defined!')


class Kosko(Inference):
    """
        Kosko's inference method.
    """
    @staticmethod    
    def infer(**kwargs) -> np.array:
        """
            Kosko's inference method.

            Parameters
            ----------
            initial_state: numpy.array
                                initial state vector of the concepts
            weight_matrix: numpy.ndarray
                            N*N weight matrix of the FCM.
            
            Return
            -------
            y: numpy.array
                    updated state vector
        """
        initial_state = kwargs['initial_state']
        weight_matrix = kwargs['weight_matrix']

        weight_matrix = weight_matrix.T
        res = weight_matrix.dot(initial_state)
        
        return res


class ModifiedKosko(Inference):
    """
        Modified Kosko inference method.
    """
    @staticmethod    
    def infer(**kwargs) -> np.array:
        """
            Modified Kosko inference method.

            Parameters
            ----------
            initial_state: numpy.array
                            initial state vector of the concepts
            weight_matrix: numpy.ndarray
                            N*N weight matrix of the FCM.

            Return
            ----------
            y: numpy.array
                    updated state vector
        """
        initial_state = kwargs['initial_state']
        weight_matrix = kwargs['weight_matrix']

        weight_matrix = weight_matrix.T
        res = weight_matrix.dot(initial_state) + initial_state
        
        return res


class Rescaled(Inference):
    """
        Rescaled inference method.
    """
    @staticmethod    
    def infer(**kwargs) -> np.array:
        """
            Rescaled inference method.

            Parameters
            ----------
            initial_state: numpy.array
                            initial state vector of the concepts
            weight_matrix: numpy.ndarray
                            N*N weight matrix of the FCM.
            
            Return
            ----------
            y: numpy.array
                    updated state vector
        """
        initial_state = kwargs['initial_state']
        weight_matrix = kwargs['weight_matrix']
        weight_matrix = weight_matrix.T

        res = weight_matrix.dot(([2*i-1 for i in initial_state])) + ([2*i-1 for i in initial_state])
        
        return res


###########################################################################
##                       RCGA fro FCMs                                   ##
###########################################################################
import pandas as pd
from fcmpy.expert_fcm.input_validator import type_check
from fcmpy.ml.genetic.ga_interface import GA
from fcmpy.store.methodsStore import GetMethod
from fcmpy.ml.genetic.generational_ga import GRCGA
from fcmpy.ml.genetic.ssga import SSGA


# Factory of genetic algorithms
class RcgaStore(GetMethod):
    """
        RCGA algorithms for learning FCMs.
    """
    __methods = {'generational':GRCGA, 'ssga' : SSGA}

    @staticmethod
    @type_check
    def get(method:str):
        """
            Get the respective RCGA method based on user input.
        """
        if method in RcgaStore.__methods.keys():
            return RcgaStore.__methods[method]
        else:
            raise ValueError('The RCGA type is not defined.')


class RCGA(GA):
    """
        Real Coded Genetic Algorithm (RCGA) for learning FCM connection matrices.

        Parameters
        ----------
        data: pd.DataFrame

        ga_type: str
                    type of genetic algorithm --> "generational", "ssga"
                    default --> "generational"
        
        init_type: str
                    type of initialization
                    default --> "uniform"
        
        population_size: int
                            number of chromosome to be generated in the population
                            default --> 100
        
        fitness_type: str
                        fitness function
                        default -> "stach_fitness"
        
        normalization_type: str
                            type of normalization to apply onto the matrix error
                            normalization type --> "L1", "L2", "LInf"
                            default --> L2
        
        a: int
            parameter for the fitness function
            default -> 100
        
        p: int
            parameter for the fitness function
            default --> 2

        inference: str
                    inference method for the FCM update
                    inference method --> "kosko", "mKosko", "rescaled"
                    default -> 'mKosko'
        
        transfer: str
                    transfer function to be applied on the updated concept values
                    transfer function --> "sigmoid", "bivalent", "trivalent", "tanh"
                    default --> "sigmoid"
        
        l: int
            slope parameter for the sigmoid transfer function
            default --> 1        
    """
    def __init__(self, data:pd.DataFrame, ga_type:str='generational', init_type:str='uniform', population_size:int=100,
                    fitness_type:str='stach_fitness', normalization_type:str='L2', a:int=100, p:int=2,
                    inference:str='mKosko', transfer:str='sigmoid', l:int=1, **kwargs):
                
        self.__init_type = init_type
        self.__popSize = population_size
        self.__data = data
        self.__fitness_type = fitness_type
        self.__normalization_type = normalization_type
        self.__a = a
        self.__p = p
        self.__inference = inference
        self.__transfer = transfer
        self.__l=l
        self.__rcga = RcgaStore.get(ga_type)
        self.__kwargs = kwargs
        self.fitness = None
        self.solution = None

    def run(self, n_iterations:int=100000, n_participants:int=5, recombination_type:str='one_point_crossover',
                p_recombination:float=0.9, p_mutation:float=0.5, b:int=5, threshold:float=0.9, **params): 
        """
            Run the RCGA learning algorithm for learning FCM connection matrices

            Parameters
            ----------
            n_iterations: int
                            number of iterations to run
                            default -> 100000
            
            n_participants: int
                            number of participants in the selection process
            
            recombination_type: str
                                recombination type to use
                                default -> "one_point_crossover"
            
            p_recombination: float
                                probability of recombination
                                default -> 0.9
            
            p_mutation: float
                        probability of mutation
                        default -> 0.5
            
            b: int
                parameter for the fitness function
                default -> 5
            
            threshold: float
                        threshold of acceptable fitness
            
            Return
            ------
            y: pd.DataFrame
                optimized connection matrix
        """
        rcga = self.__rcga(data=self.__data, init_type=self.__init_type, population_size=self.__popSize,
                            fitness_type=self.__fitness_type, normalization_type=self.__normalization_type, a=self.__a, p=self.__p,
                            inference=self.__inference, transfer=self.__transfer, l=self.__l, kwargs=self.__kwargs)

        res = rcga.run(n_iterations=n_iterations, n_participants=n_participants, recombination_type=recombination_type,
                            p_recombination=p_recombination, p_mutation=p_mutation, b=b, threshold=threshold, params=params)  

        df = pd.DataFrame(res['solution'], columns=self.__data.keys(), index=self.__data.keys())
        
        self.fitness = res['fitness']
        self.solution = df



import pandas as pd
import numpy as np
import warnings
from abc import ABC
from abc import abstractmethod
from typing import Union
from tqdm import tqdm
from fcmpy.expert_fcm.input_validator import type_check
from .update_weight import NhlWeightUpdate
from .update_state import FcmUpdate
from .termination import FirstCriterion
from .termination import SecondCriterion


class HebbianLearning(ABC):
    """
        Hebbian Based Learning for FCMs.
    """
    @abstractmethod
    def run(**kwargs) -> pd.DataFrame:
        raise NotImplementedError('run method is not defined.')


class NHL(HebbianLearning):
    """
        NHL algorithm for optimizing the FCM weight matrix.
    """
    @type_check
    def __init__(self, state_vector: dict, weight_matrix:Union[np.ndarray, pd.DataFrame], 
                    doc_values:dict) -> None:
        """
            NHL algorithm for optimizing the FCM weight matrix.

            Parameters
            ----------
            state_vector : dict
                            state vector of the concepts
                            keys ---> concepts, values ---> value of the associated concept
            
            weight_matrix : pd.DataFrame
                            N*N weight matrix of the FCM. 
            
            doc_values : dict
                            DOC values where the keys are the DOCs and 
                            the values are lists with min max values of the DOCs.
        """
        self.__state_vector = state_vector
        self.__weight_matrix = weight_matrix
        self.__doc_values = doc_values
        self.__weightUpdate = NhlWeightUpdate()
        self.__stateUpdate = FcmUpdate()
        self.__termination1 = FirstCriterion()
        self.__termination2 = SecondCriterion()

    @type_check
    def run(self, decay=1, learning_rate=0.01, iterations:int= 100, transfer:str= 'sigmoid', 
                    inference:str='mKosko', thresh:float = 0.002, l:Union[float, int]=0.98, **kwargs) -> pd.DataFrame:
        """
            Run the NHL algorithm.

            Parameters
            ----------
            decay: float,
                    default --> 1
            
            learning_rate: int/float
                            default --> 0.01
            
            iterations: int
                        iterations for the NHL algorithm to run
                        default --> 100

            transfer: str
                        transfer function --> "sigmoid", "bivalent", "trivalent", "tanh"
                        default -> "sigmoid"

            l: float
                parameter for the sigmoid transfer function

            inference: str
                        inference method --> "kosko", "mKosko", "rescaled"

            thresh: float
                        threshold of error for the F2 termination condition for NHL algorithm
                        default --> 0.002

            Return
            -------
            y : pd.DataFrame
                    the optimized weight matrix
        """
        gamma = decay
        eta = learning_rate
        # Initialize the prior state vector and weight matrices.
        s_prior = self.__state_vector.copy()
        w_prior = self.__weight_matrix.copy()

        for _ in tqdm(range(iterations)):
            w_new = self.__weightUpdate.update(initial_state=s_prior, weight_matrix=w_prior,
                                                    gamma=gamma, eta=eta)

            s_new = self.__stateUpdate.update(state_vector=s_prior, weight_matrix=w_new, transfer=transfer,
                                                inference=inference, l=l, kwargs=kwargs) 
            
            if self.__termination1.terminate(doc_values=self.__doc_values, state_vector_prior=s_prior, state_vector_current=s_new) and \
                                                self.__termination2.terminate(doc_values=self.__doc_values, state_vector_prior=s_prior, 
                                                                    state_vector_current=s_new, thresh=thresh):

                print(f'The NHL learning process converged at step {_} with the learning rate eta = {eta} and decay = {gamma}!')
                return w_new
            else:
                s_prior = s_new
                w_prior = w_new
        
        if _ >= iterations-1:
            warnings.warn('The NHL did not converge! Consider a different set of parameters.')
            return w_new


import pandas as pd
import collections
from typing import Union
from fcmpy.expert_fcm.input_validator import type_check


class Transform:
    """
        Collection of data transformation (static) methods.
    """
    @staticmethod
    @type_check
    def calculateProportions(data: pd.DataFrame, conceptPair: tuple, nExperts:int) -> dict:
        """
            Calculate the proportions of answers to each linguistic term.

            Parameters
            ----------
            data: pandas.DataFrame
                        Data frame that contains all the expert 
                        inputs (i.e., "flattened" OrderedDict)
            
            conceptPair: tuple,
                            concept pair for which the activation parameter
                            should be constructed

            Return
            -------
            y: dict,
                keys ---> linguistic terms, values ---> proportion
                of expert ratings.
        """
        activation_parameter = {}
        activation_parameter = (data.loc[conceptPair].sum()/nExperts).to_dict()

        return activation_parameter

    @staticmethod
    @type_check
    def flatData(data: Union[dict, collections.OrderedDict]) -> pd.DataFrame:
        """
            Create a flat data from an ordered dictionary.

            Parameters
            ----------
            data: dict,
                    keys ---> expertId, values ---> pandas.DataFrame
            
            Return
            -------
            y: pandas.DataFrame
                data with all the expert inputs in one dataframe.
        """
        # Create a flat data with all of the experts' inputs.
        flat_data = pd.concat([data[i] for i in data], sort = False)
        flat_data.columns = [i.lower() for i in flat_data.columns]
        flat_data = flat_data.set_index(['from', 'to'])
        flat_data = flat_data.sort_index(level=['from','to'])

        return flat_data



import numpy as np
from abc import ABC
from abc import abstractmethod
from fcmpy.expert_fcm.input_validator import type_check

class Termination(ABC):
    """
        Interface for terminating the learning algorithms
    """
    @abstractmethod
    def terminate() -> bool:
        """
            Determine whether to stop the learning process.
        """
        raise NotImplementedError('Terminate method is not defined!')


class FirstCriterion(Termination):
    """
        Termination method based on F1 NHL
    """
    @staticmethod
    @type_check
    def __sumOfSquared(doc_values:dict, state_vector: dict) -> float:
        """
            Calculate the sum of the squared errors between the doc at step K and step k+1.
            
            Parameters
            ----------
            doc_values: dict
                        Desired output concepts (DOC)
                        keys ---> concepts, values ---> list of maximum and minimum values of the DOCs

            state_vector: dict
                            State vector of the concepts
                            keys ---> concepts, values ---> State of the associated concept

            Return
            -------
            y: float
                sum of the squared errors
        """
        l = []
        for i in doc_values.keys():
            t = sum(doc_values[i])/len(doc_values[i])
            l.append(state_vector[i] - t)
        res = np.sqrt(sum([i**2 for i in l]))
        return res

    @staticmethod
    @type_check
    def __checkRange(doc_values:dict, state_vector:dict) -> bool:
        """
            Check if the doc values are in the desired range.
        
            Parameters
            ----------
            doc_values: dict
                        Desired output concepts (DOC)
                        keys ---> concepts, values ---> list of maximum and minimum values of the DOCs

            state_vector: dict
                            State vector of the concepts
                            keys ---> concepts, values ---> State of the associated concept

            Return
            -------
            y: Bool
                True if the DOCs are in the desired range, False if otherwise.
        """
        vals = dict((key, state_vector[key]) for key in doc_values.keys())
        ps = []
        for i in vals.keys():
            if doc_values[i][0] <= vals[i] <= doc_values[i][1]:
                ps.append(True)
        if len(ps) == len(doc_values.keys()):
            return True

    @staticmethod
    @type_check
    def terminate(doc_values:dict, state_vector_prior: dict, state_vector_current: dict) -> bool:
        """
            The objective of the training process is to find the set of
            weights that minimize function F1.
        
            Parameters
            ----------
            doc_values: dict
                            DOC values where the keys are the DOCs and 
                            the values are lists with min max values of the DOCs.

            state_vector_prior: np.ndarray
                            state vector at time t-1
            
            state_vector_current: np.ndarray
                            state vector at time t

            Return
            -------
            y: bool
                True if the condition is satisfied, False if otherwise.
        """
        prior = FirstCriterion.__sumOfSquared(doc_values=doc_values, state_vector=state_vector_prior)
        current = FirstCriterion.__sumOfSquared(doc_values=doc_values, state_vector=state_vector_current)

        if (current < prior) and FirstCriterion.__checkRange(doc_values=doc_values, state_vector=state_vector_current):
            return True
        else:
            return False


class SecondCriterion(Termination):
    """
        Termination method based on F2
    """
    @staticmethod
    def __absDifference(doc_values:dict, state_vector_prior: dict, state_vector_current: dict):
        """
            Calculate the maximum absolute difference between the desired output concepts (DOCs) at step K
            and step K+1.
        
            Parameters
            ----------
            doc_values: dict
                            DOC values where the keys are the DOCs and 
                            the values are lists with min max values of the DOCs.

            state_vector_prior: np.ndarray
                            state vector at time t-1
            
            state_vector_current: np.ndarray
                            state vector at time t

            Return
            -------
            y: float
                maximum absolute difference between the prior and current state values
        """
        prior = dict((key, state_vector_prior[key]) for key in doc_values.keys())
        current = dict((key, state_vector_current[key]) for key in doc_values.keys())
        
        dif = {key: abs(current[key] - prior.get(key, 0)) for key in prior}.values()
        max_dif = max(list(dif))
        
        return max_dif

    @staticmethod
    def terminate(doc_values:dict, state_vector_prior: dict, 
                        state_vector_current: dict, thresh:float=0.002) -> bool:
        """
            Calculate the termination condition based on the absolute difference.
        
            Parameters
            ----------
            doc_values: dict
                            DOC values where the keys are the DOCs and 
                            the values are lists with min max values of the DOCs.

            state_vector_prior: np.ndarray
                            state vector at time t-1
            
            state_vector_current: np.ndarray
                            state vector at time t
            
            thresh: float
                        threshold of absolute difference between the
                        prior and current values of the states

            Return
            -------
            y: bool
                True if the maximum difference between the steps is < the threshold,
                False if otherwise.
        """
        max_dif = SecondCriterion.__absDifference(doc_values, state_vector_prior, state_vector_current)
        
        if max_dif < thresh:
            return True
        else:
            return False